#include <stdint.h>
#include <inttypes.h>
#include <time.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include <sys/resource.h>

#include "list.h"
#include "util.h"
#include "tgtd.h"
#include "work.h"
#include "driver.h"
#include "target.h"

#define NS_PER_SECOND 1000000000ULL

static const bool lat_direction_read[CMDCNT_LIMIT] = {
	[CMDCNT_READOP_QRLAT] = true,
	[CMDCNT_READOP_DEVLAT] = true,
	[CMDCNT_READOP_TXLAT] = true,
};

struct cmd_life_range {enum cmd_life_event start, end;};

static const struct cmd_life_range lat_source[CMDCNT_LIMIT] = {
	[CMDCNT_READOP_QRLAT] = {CMDLIFE_QUEUED, CMDLIFE_RUNNING},
	[CMDCNT_READOP_DEVLAT] = {CMDLIFE_RUNNING, CMDLIFE_DONE_DEVICE},
	[CMDCNT_READOP_TXLAT] = {CMDLIFE_DONE_DEVICE, CMDLIFE_DONE_TX},
	[CMDCNT_WRITEOP_RXLAT] = {CMDLIFE_RX_START, CMDLIFE_QUEUED},
	[CMDCNT_WRITEOP_QRLAT] = {CMDLIFE_QUEUED, CMDLIFE_RUNNING},
	[CMDCNT_WRITEOP_DEVLAT] = {CMDLIFE_RUNNING, CMDLIFE_DONE_DEVICE},
};

void ts_invalidate(struct timespec *ts) {
	ts->tv_sec = 0;
	ts->tv_nsec = 0;
}

bool ts_is_valid(const struct timespec *ts)
{
	return (ts->tv_sec!=0);
}

uint64_t ts_valid_diff(const struct timespec *start,
		       const struct timespec *end)
{
	if (ts_is_valid(start)&&ts_is_valid(end)) {
		uint64_t result = ((end->tv_sec-start->tv_sec)*NS_PER_SECOND)+
			(uint64_t)(end->tv_nsec-start->tv_nsec);
		/* Zero elapsed time is clamped to 1ns elapsed time
		 * (0ns is semantically "unknown interval"). Very
		 * large unsigned value of elapsed time means time's
		 * gone back, the interval is clamped to 1ns again. */
		if (result==0 || result > ((uint64_t)-NS_PER_SECOND))
			return 1;
		else
			return result;
	} else {
		return 0;
	}
}

uint32_t tv_diff_ms(const struct timeval *start,const struct timeval *end)
{
	return ((end->tv_usec-start->tv_usec)+(end->tv_sec-start->tv_sec)*1000000ULL)/1000ULL;
}

void cmd_account_latency(struct scsi_cmd* cmd, enum cmd_latency_type lat, struct stat_entry *se) {
	enum data_direction cmd_dir = scsi_get_data_dir(cmd);
	uint64_t elapsed;
	if (cmd_dir!=DATA_BIDIRECTIONAL) {
		if (lat_direction_read[lat]!=(cmd_dir==DATA_READ)) {
			return;
		}
	}
	elapsed = ts_valid_diff(&cmd->life_events.ts[lat_source[lat].start],
				&cmd->life_events.ts[lat_source[lat].end]);
	if (elapsed) {
		++se->lat[lat].ncommands;
		se->lat[lat].latsum += elapsed;
		if (se->lat[lat].latmax < elapsed)
			se->lat[lat].latmax = elapsed;
	}
}

void cmd_account_xfer(struct scsi_cmd* cmd, enum cmd_direction dir, struct stat_entry *se) {
	enum data_direction cmd_dir = scsi_get_data_dir(cmd);
	if (cmd_dir!=DATA_BIDIRECTIONAL) {
		if ((dir==CMDDIR_READ)!=(cmd_dir==DATA_READ)) {
			return;
		}
	}

	++se->xfer[dir].ncommands;
	se->xfer[dir].total_length += scsi_get_in_length(cmd) +
		scsi_get_out_length(cmd);
}

void cmd_update_lun_stat(struct scsi_cmd* cmd) {
	int i;
	struct scsi_lu *lu = cmd->dev;
	struct stat_entry *se = &lu->stath.ent[lu->stath.last_nbuf % CMDSTAT_NBUF];
	for (i=0; i<CMDCNT_LIMIT; ++i) {
		cmd_account_latency(cmd,i,se);
	}
	for (i=0; i<CMDDIR_LIMIT; ++i) {
		cmd_account_xfer(cmd,i,se);
	}
}

void cmd_track_time(struct scsi_cmd* cmd, enum cmd_life_event what) {
	if (clock_gettime(CLOCK_MONOTONIC,&cmd->life_events.ts[what])) {
		ts_invalidate(&cmd->life_events.ts[what]);
	} else {
		if (what == CMDLIFE_LIMIT-1) {
			cmd_update_lun_stat(cmd);
		}
	}
}

void dev_track_queue(struct scsi_lu *lu, enum dev_queue what, uint32_t depth) {
	struct stat_queue *q = &lu->stath.ent[lu->stath.last_nbuf % CMDSTAT_NBUF].queue[what];
	q->nentries = 1;
	q->depthsum = q->depthmax = depth;
}

void cmd_stat_add(struct stat_entry *total, struct stat_entry *delta) {
	int i;
	for (i=0; i<CMDCNT_LIMIT; ++i) {
		total->lat[i].ncommands+=delta->lat[i].ncommands;
		total->lat[i].latsum+=delta->lat[i].latsum;
		if (total->lat[i].latmax<delta->lat[i].latmax)
			total->lat[i].latmax=delta->lat[i].latmax;
	}
	for (i=0; i<DEVQUEUE_LIMIT; ++i) {
		total->queue[i].nentries+=delta->queue[i].nentries;
		total->queue[i].depthsum+=delta->queue[i].depthsum;
		if (total->queue[i].depthmax<delta->queue[i].depthmax)
			total->queue[i].depthmax=delta->queue[i].depthmax;
	}
	for (i=0; i<CMDDIR_LIMIT; ++i) {
		total->xfer[i].ncommands+=delta->xfer[i].ncommands;
		total->xfer[i].total_length+=delta->xfer[i].total_length;
	}
	for (i=0; i<PROCUSAGE_LIMIT; ++i) {
		total->usage[i].nentries+=delta->usage[i].nentries;
		total->usage[i].valuesum+=delta->usage[i].valuesum;
		if (total->usage[i].valuemax<delta->usage[i].valuemax)
			total->usage[i].valuemax=delta->usage[i].valuemax;
	}
}

/*
  n entries at totals[] are filled with aggregate statistics, each ith
  entry being the "sum" of nsamples[i] latest entries. The function
  assumes nsamples[i+1]>nsamples[i] \forall i \in [0;n-1)
*/
void cmd_stat_aggregate(struct stat_entry totals[], struct stat_history *stath, const int *nsamples, int n)
{
	int i,j,k;
	uint32_t cnbuf;

	cnbuf = stath->last_nbuf;

	for (i=0; i<n; ++i) {
		if (i) {
			totals[i] = totals[i-1];
		} else {
			memset(&totals[i],0,sizeof totals[i]);
		}
		k = nsamples[i]-(i?nsamples[i-1]:0);
		for (j=0;j<k;++j) {
			--cnbuf;
			cmd_stat_add(&totals[i],&stath->ent[cnbuf % CMDSTAT_NBUF]);
		}
	}
}

static const int report_nsamples[] = {
	1, 				/* 5sec */
	5*60/CMDSTAT_SAMPLE_SECONDS,	/* 5min */
	15*60/CMDSTAT_SAMPLE_SECONDS,	/* 15min */
};

static const char* report_spans[] = {
	"5 sec",
	"5 min",
	"15 min",
};

static void latency_print_json(struct latency_stat *latency, const char *label, struct concat_buf *b)
{
	concat_printf(b, "\"%s\" : {", label);

	const char *label_src[] = { "\"span1ms\" : [", "], \"span10ms\" : [", "], \"span100ms\" : [" };
	uint32_t *data_src[] = { latency->span1ms, latency->span10ms, latency->span100ms };

	unsigned src_idx;
	for (src_idx = 0; src_idx < sizeof(data_src) / sizeof(uint32_t *); src_idx++) {
		concat_printf(b, label_src[src_idx]);
		unsigned data_idx;
		for (data_idx = 0; data_idx < PERF_PROFILE_SIZE; data_idx++) {
			if (data_idx != 0)
				concat_printf(b, ", ");
			concat_printf(b, "%" PRIu32, data_src[src_idx][data_idx]);
		}
	}

	concat_printf(b, "]}");
}

static void add_lu_stat(struct lu_stat *dst, struct lu_stat *src)
{
	dst->err_num += src->err_num;
	dst->aborted_num += src->aborted_num;
	dst->bidir_done_cmds += src->bidir_done_cmds;
	dst->bidir_subm_cmds += src->bidir_subm_cmds;
	dst->rd_done_bytes += src->rd_done_bytes;
	dst->rd_done_cmds += src->rd_done_cmds;
	dst->rd_subm_bytes += src->rd_subm_bytes;
	dst->rd_subm_cmds += src->rd_subm_cmds;
	dst->wr_done_bytes += src->wr_done_bytes;
	dst->wr_done_cmds += src->wr_done_cmds;
	dst->wr_subm_bytes += src->wr_subm_bytes;
	dst->wr_subm_cmds += src->wr_subm_cmds;

	struct sum_work_t {
		uint32_t *dst;
		uint32_t *src;
	} sum_work[] = {
		{ dst->read_latency.span1ms, src->read_latency.span1ms },
		{ dst->read_latency.span10ms, src->read_latency.span10ms },
		{ dst->read_latency.span100ms, src->read_latency.span100ms },
		{ dst->write_latency.span1ms, src->write_latency.span1ms },
		{ dst->write_latency.span10ms, src->write_latency.span10ms },
		{ dst->write_latency.span100ms, src->write_latency.span100ms },
		{ dst->sync_latency.span1ms, src->sync_latency.span1ms },
		{ dst->sync_latency.span10ms, src->sync_latency.span10ms },
		{ dst->sync_latency.span100ms, src->sync_latency.span100ms },
	};

	int work_idx, idx;
	for (work_idx = 0; work_idx < sizeof(sum_work) / sizeof(struct sum_work_t); work_idx++)
		for (idx = 0; idx < PERF_PROFILE_SIZE; idx++)
			sum_work[work_idx].dst[idx] += sum_work[work_idx].src[idx];
}

static void device_total_print_json(struct concat_buf *b, struct scsi_lu *dev)
{
	struct it_nexus_lu_info *itn_lu;

	struct lu_stat cumulative_stat;
	memset(&cumulative_stat, 0, sizeof(struct lu_stat));
	uint32_t initiators = 0;
	list_for_each_entry(itn_lu, &dev->lu_itl_info_list, lu_itl_info_siblings) {
		initiators++;
		add_lu_stat(&cumulative_stat, &itn_lu->stat);
	}

	concat_printf(b,
		"\"initiators\" : %" PRIu32 ","
		"\"rd_subm_bytes\" : %" PRIu64 ", \"rd_subm_cmds\" : %" PRIu32 ","
		"\"rd_done_bytes\" : %" PRIu64 ", \"rd_done_cmds\" : %" PRIu32 ","
		"\"wr_subm_bytes\" : %" PRIu64 ", \"wr_subm_cmds\" : %" PRIu32 ","
		"\"wr_done_bytes\" : %" PRIu64 ", \"wr_done_cmds\" : %" PRIu32 ","
		"\"errs\" : %" PRIu32 ", \"aborted\" : %" PRIu32 ",",
		initiators,
		cumulative_stat.rd_subm_bytes, cumulative_stat.rd_subm_cmds,
		cumulative_stat.rd_done_bytes, cumulative_stat.rd_done_cmds,
		cumulative_stat.wr_subm_bytes, cumulative_stat.wr_subm_cmds,
		cumulative_stat.wr_done_bytes, cumulative_stat.wr_done_cmds,
		cumulative_stat.err_num, cumulative_stat.aborted_num);

	latency_print_json(&cumulative_stat.read_latency, "read_latency", b);
	concat_printf(b, ",");
	latency_print_json(&cumulative_stat.write_latency, "write_latency", b);
	concat_printf(b, ",");
	latency_print_json(&cumulative_stat.sync_latency, "sync_latency", b);
}

static void cmdstat_device_print_json(struct concat_buf *b, struct scsi_lu *dev)
{
	struct stat_entry totals[ARRAY_SIZE(report_nsamples)];
	cmd_stat_aggregate(totals,&dev->stath,report_nsamples,ARRAY_SIZE(report_nsamples));
	concat_printf(b, "\"timings\" : ["); // open JSON
	int rep_nsmpl_idx;
	for (rep_nsmpl_idx=0; rep_nsmpl_idx<ARRAY_SIZE(report_nsamples); ++rep_nsmpl_idx) {
		concat_printf(b,"{\"interval\" : \"%s\",", report_spans[rep_nsmpl_idx]);

		const char *col_head[]={"lat_recv","lat_queue","lat_dev","lat_send"};
		const char *dirs[] = {"read", "write"};
		int col_head_idx, dir_idx;
		for (col_head_idx = 0; col_head_idx < ARRAY_SIZE(col_head); ++col_head_idx) {
			concat_printf(b, "\"%s\" : {", col_head[col_head_idx]);

			int first_dir_entry = 1;
			for (dir_idx = 0; dir_idx < ARRAY_SIZE(dirs); ++dir_idx) {
				int base = (dir_idx ? CMDCNT_WRITEOP_RXLAT : CMDCNT_READOP_QRLAT);
				int dat_offset = ((col_head_idx + dir_idx) % ARRAY_SIZE(col_head)) - 1;

				struct stat_latency *sl = (dat_offset >= 0) ? &totals[rep_nsmpl_idx].lat[base+dat_offset] : NULL;
				if (!sl)
					continue;

				concat_printf(b, "%s\"%s\" : {", (dir_idx && !first_dir_entry) ? "," : "", dirs[dir_idx]);
				first_dir_entry = 0;

				const char *sub[] = { "avg", "max" };
				int sb_dir_idx;
				for (sb_dir_idx = 0; sb_dir_idx < ARRAY_SIZE(sub); ++sb_dir_idx) {
					if (sl->ncommands) {
						uint64_t r =  sb_dir_idx ? sl->latmax : sl->latsum / sl->ncommands;
						concat_printf(b, "\"%s\" : %d.%d", sub[sb_dir_idx], (int)(r/1000000), (int)((r%1000000)/1000));
					} else {
						concat_printf(b, "\"%s\" : \"\"", sub[sb_dir_idx]);
					}
					concat_printf(b, "%s", (sb_dir_idx < ARRAY_SIZE(sub) - 1) ? "," : "");
				}

				concat_printf(b, "}");
			}
			concat_printf(b, "},");
		}

		for (dir_idx = 0; dir_idx < ARRAY_SIZE(dirs); ++dir_idx) {
			concat_printf(b, "\"%s_total\" : {", dirs[dir_idx]);
			concat_printf(b, "\"commands\" : %"PRIu32",", totals[rep_nsmpl_idx].xfer[dir_idx].ncommands);
			concat_printf(b, "\"bytes\" : %"PRIu64, totals[rep_nsmpl_idx].xfer[dir_idx].total_length);
			concat_printf(b, "},");
		}

		int j;
		const char *queues[] = {"waiting","in_flight"};
		concat_printf(b, "\"queue_depth\" : {");
		for (j=0; j<DEVQUEUE_LIMIT; ++j) {
			if (totals[rep_nsmpl_idx].queue[j].nentries) {
				concat_printf(b, "\"%s\" : {\"avg\" : %"PRIu32", \"max\" : %"PRIu32"}%s",
					queues[j],
					totals[rep_nsmpl_idx].queue[j].depthsum/totals[rep_nsmpl_idx].queue[j].nentries,
					totals[rep_nsmpl_idx].queue[j].depthmax,
					(j < DEVQUEUE_LIMIT-1) ? "," : "");
			}
		}
		concat_printf(b, "},"); // close queue_depth

		const char *usages[] = {"rss","cpu"};
		concat_printf(b, "\"proc_res\" : {");
		for (j=0; j<PROCUSAGE_LIMIT; ++j) {
			if (totals[rep_nsmpl_idx].usage[j].nentries) {
				concat_printf(b, "\"%s\" : {\"avg\" : %"PRIu32", \"max\" : %"PRIu32"}%s",
					usages[j],
					totals[rep_nsmpl_idx].usage[j].valuesum/totals[rep_nsmpl_idx].usage[j].nentries,
					totals[rep_nsmpl_idx].usage[j].valuemax,
					(j < PROCUSAGE_LIMIT-1) ? "," : "");
			}
		}
		concat_printf(b, "}"); // close proc_res
		concat_printf(b, "}%s", (rep_nsmpl_idx < ARRAY_SIZE(report_nsamples) - 1) ? "," : ""); // close Interval
	}
	concat_printf(b, "]");
}

void device_print_json(struct target *target, struct concat_buf *b, struct scsi_lu *lu)
{
	concat_printf(b,
		"{\"lun\" : %" PRIu64 ","
		"\"reconnections\" : %" PRIu64 ",",
		lu->lun, target->stat.reconnects);

	device_total_print_json(b, lu);
	concat_printf(b, ",");
	cmdstat_device_print_json(b, lu);
	concat_printf(b, "}");
}

void cmdstat_device_print(struct concat_buf *b, struct scsi_lu *dev) {
	int i,j,k,c;
	struct stat_entry totals[ARRAY_SIZE(report_nsamples)];
	const char *col_head[]={"IOdir(fun)","lat(recv)","lat(queue)","lat(dev)","lat(send)","#cmds/bytes"};
	cmd_stat_aggregate(totals,&dev->stath,report_nsamples,ARRAY_SIZE(report_nsamples));
	for (i=0; i<ARRAY_SIZE(report_nsamples); ++i) {
		const char *dirs[]={"Read","Write"};
		const char *queues[]={"Waiting","In flight"};
		const char *usages[]={"RSS(KiB)","CPU(%)"};
		concat_printf(b,"\nLatest %s interval (latencies in ms):\n", report_spans[i]);
		for (j=0; j<6; ++j)
			concat_printf(b,"%10s%s",col_head[j],j==5?"\n":" |");
		for (j=0; j<6; ++j)
			concat_printf(b,"-----------%s",j==5?"-----------\n":"+");

		for (j=0; j<ARRAY_SIZE(dirs); ++j) {
			const char *sub[]={"Avg","Max"};
			for (k=0; k<2; ++k) {
				concat_printf(b, "%5s(%3s) |",dirs[j],sub[k]);
				if (!j) {
					concat_printf(b, "%10s |","  ---  ");
				}
				for (c=0;c<3;++c) {
					int base = (j?CMDCNT_WRITEOP_RXLAT : CMDCNT_READOP_QRLAT);
					struct stat_latency *sl = &totals[i].lat[c+base];
					if (sl->ncommands) {
						uint64_t r =  k? sl->latmax : sl->latsum / sl->ncommands;
						concat_printf(b, "%6d.%03d |", (int)(r/1000000),
							      (int)((r%1000000)/1000));
					} else {
						concat_printf(b, "%10s |", " n/a");
					}
				}
				if (j) {
					concat_printf(b, "%10s |","  ---  ");
				}
				if (!k) {
					concat_printf(b, "%"PRIu32" commands\n", totals[i].xfer[j].ncommands);
				} else {
					concat_printf(b, "%"PRIu64" bytes\n", totals[i].xfer[j].total_length);
				}
			}
		}
		concat_printf(b,"== queue depth data ===================================\n");
		for (j=0; j<DEVQUEUE_LIMIT; ++j) {
			if (totals[i].queue[j].nentries) {
				concat_printf(b, "Queue depth |%10s| average %5"PRIu32
					      " | maximum %5"PRIu32"\n",
					      queues[j],
					      totals[i].queue[j].depthsum/totals[i].queue[j].nentries,
					      totals[i].queue[j].depthmax);
			}
		}
		concat_printf(b,"== process resource usage =============================\n");
		for (j=0; j<PROCUSAGE_LIMIT; ++j) {
			if (totals[i].usage[j].nentries) {
				concat_printf(b, "%10s| average %10"PRIu32" | maximum %10"PRIu32"\n",
					      usages[j],
					      totals[i].usage[j].valuesum/totals[i].usage[j].nentries,
					      totals[i].usage[j].valuemax);
			}
		}
		concat_printf(b,"=======================================================\n");
	}
}

static struct tgt_work work;

static struct rusage usage;

void cmdstat_work_handler(void* arg) {
	int i;
	struct target *tgt;
	struct scsi_lu *dev;
	struct rusage new_usage;
	uint32_t cpu_pc = 0;
	uint32_t rss_mb = 0;
	int rss_pg;
	int have_rusage = 0, have_rss = 0;
	FILE* fp = fopen("/proc/self/statm","r");
	if (!getrusage(RUSAGE_SELF,&new_usage)) {
		have_rusage = 1;
		cpu_pc = tv_diff_ms(&usage.ru_utime,&new_usage.ru_utime)*100/
			(1000*CMDSTAT_SAMPLE_SECONDS);
		usage = new_usage;
	}
	if (fp) {
		if (fscanf(fp,"%*d %d",&rss_pg)==1) {
			have_rss = 1;
			rss_mb = rss_pg*sysconf(_SC_PAGE_SIZE)/1024;
		}
		fclose(fp);
	}
	for (i = 0; tgt_drivers[i]; i++) {
		list_for_each_entry(tgt,&tgt_drivers[i]->target_list, lld_siblings) {
			list_for_each_entry(dev,&tgt->device_list, device_siblings) {
				struct stat_usage *u =
					dev->stath.ent[dev->stath.last_nbuf % CMDSTAT_NBUF].usage;
				if (have_rusage) {
					u[PROCUSAGE_CPUTIME].valuesum =
						u[PROCUSAGE_CPUTIME].valuemax = cpu_pc;
					u[PROCUSAGE_CPUTIME].nentries=1;
				}
				if (have_rss) {
					u[PROCUSAGE_MAXRSS].valuesum =
						u[PROCUSAGE_MAXRSS].valuemax = rss_mb;
					u[PROCUSAGE_MAXRSS].nentries=1;
				}
				if (dev->bst->bs_stat_queue)
					dev->bst->bs_stat_queue(dev);
				++dev->stath.last_nbuf;
				memset(&dev->stath.ent[dev->stath.last_nbuf % CMDSTAT_NBUF],
				       0,sizeof dev->stath.ent[0]);
			}
		}
	}
	add_work(&work,CMDSTAT_SAMPLE_SECONDS);
}

void cmdstat_alldev_print(struct concat_buf *b) {
	int i;
	struct target *tgt;
	struct scsi_lu *dev;
	for (i = 0; tgt_drivers[i]; i++) {
		list_for_each_entry(tgt,&tgt_drivers[i]->target_list, lld_siblings) {
			list_for_each_entry(dev,&tgt->device_list, device_siblings) {
				if (dev->lun) {
					concat_printf(b, "Target: %s LUN: %"PRIx64"\n",
						      dev->tgt->name,
						      dev->lun);
					cmdstat_device_print(b,dev);
				}
			}
		}
	}
}

void cmdstat_init(void) {
	getrusage(RUSAGE_SELF,&usage);
	add_work(&work,CMDSTAT_SAMPLE_SECONDS);
}

static struct tgt_work work = {.func = cmdstat_work_handler};
