/*
 * Copyright (C) 2002-2003 Ardis Technolgies <roman@ardistech.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2 of the
 * License.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include "memdebug.h"
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <errno.h>
#include <sys/shm.h>
#include <sys/ipc.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/prctl.h>

#include "log.h"
#include "xlog.h"

#define LOGDBG 0

#if LOGDBG
#define logdbg(file, fmt, args...) fprintf(file, fmt, ##args)
#else
#define logdbg(file, fmt, args...) do {} while (0)
#endif

static struct logarea *la;
static char *log_name;
int is_debug = 0;
static int use_syslog = 1, use_pcslog = 0;
static pid_t pid;

static int logarea_init (int size)
{
	int shmid;
	extern char mgmt_path[];
	key_t semkey;

	logdbg(stderr,"enter logarea_init\n");

	if ((shmid = shmget(IPC_PRIVATE, sizeof(struct logarea),
			    0644 | IPC_CREAT | IPC_EXCL)) == -1) {
		syslog(LOG_ERR, "shmget logarea failed %d", errno);
		return 1;
	}

	la = shmat(shmid, NULL, 0);
	if (!la) {
		syslog(LOG_ERR, "shmat logarea failed %d", errno);
		return 1;
	}

	shmctl(shmid, IPC_RMID, NULL);

	if (size < MAX_MSG_SIZE)
		size = LOG_SPACE_SIZE;

	if ((shmid = shmget(IPC_PRIVATE, size,
			    0644 | IPC_CREAT | IPC_EXCL)) == -1) {
		syslog(LOG_ERR, "shmget msg failed %d", errno);
		shmdt(la);
		return 1;
	}

	la->start = shmat(shmid, NULL, 0);
	if (!la->start) {
		syslog(LOG_ERR, "shmat msg failed %d", errno);
		shmdt(la);
		return 1;
	}
	memset(la->start, 0, size);

	shmctl(shmid, IPC_RMID, NULL);

	la->empty = 1;
	la->end = la->start + size;
	la->head = la->start;
	la->tail = la->start;

	if ((shmid = shmget(IPC_PRIVATE, MAX_MSG_SIZE + sizeof(struct logmsg),
			    0644 | IPC_CREAT | IPC_EXCL)) == -1) {
		syslog(LOG_ERR, "shmget logmsg failed %d", errno);
		shmdt(la->start);
		shmdt(la);
		return 1;
	}
	la->buff = shmat(shmid, NULL, 0);
	if (!la->buff) {
		syslog(LOG_ERR, "shmat logmsgfailed %d", errno);
		shmdt(la->start);
		shmdt(la);
		return 1;
	}

	shmctl(shmid, IPC_RMID, NULL);

	if ((semkey = ftok(mgmt_path, 'a')) < 0) {
		syslog(LOG_ERR, "semkey failed %d", errno);
		return 1;
	}

        if ((semget(semkey, 0, IPC_EXCL)) > 0) {
		/* Semkey may exists after SIGKILL tgtd */
		syslog(LOG_WARNING, "semkey 0x%x already exists", semkey);
	}
	if ((la->semid = semget(semkey, 1, 0666 | IPC_CREAT)) < 0) {
		syslog(LOG_ERR, "semget failed %d", errno);
		shmdt(la->buff);
		shmdt(la->start);
		shmdt(la);
		return 1;
	}
	syslog(LOG_INFO, "semkey 0x%x", semkey);

	la->semarg.val=1;
	if (semctl(la->semid, 0, SETVAL, la->semarg) < 0) {
		syslog(LOG_ERR, "semctl failed %d", errno);
		shmdt(la->buff);
		shmdt(la->start);
		shmdt(la);
		return 1;
	}

	return 0;
}

static void free_logarea (void)
{
	if (!la)
		return;
	semctl(la->semid, 0, IPC_RMID, la->semarg);
	shmdt(la->buff);
	shmdt(la->start);
	shmdt(la);
	la = NULL;
}

#if LOGDBG
static void dump_logarea (void)
{
	struct logmsg * msg;

	logdbg(stderr, "\n==== area: start addr = %p, end addr = %p ====\n",
		la->start, la->end);
	logdbg(stderr, "|addr     |next     |prio|msg\n");

	for (msg = (struct logmsg *)la->head; (void *)msg != la->tail;
	     msg = msg->next)
		logdbg(stderr, "|%p |%p |%i   |%s\n", (void *)msg, msg->next,
				msg->prio, (char *)&msg->str);

	logdbg(stderr, "|%p |%p |%i   |%s\n", (void *)msg, msg->next,
			msg->prio, (char *)&msg->str);

	logdbg(stderr, "\n\n");
}
#endif

static int log_enqueue(int prio, const char *fmt, va_list ap)
{
	int len, fwd;
	char buff[MAX_MSG_SIZE];
	struct logmsg * msg;
	struct logmsg * lastmsg;

	lastmsg = (struct logmsg *)la->tail;

	if (!la->empty) {
		fwd = sizeof(struct logmsg) +
		      strlen((char *)&lastmsg->str) * sizeof(char) + 1;
		la->tail += fwd;
	}
	vsnprintf(buff, MAX_MSG_SIZE, fmt, ap);
	len = strlen(buff) * sizeof(char) + 1;

	/* not enough space on tail : rewind */
	if (la->head <= la->tail &&
	    (len + sizeof(struct logmsg)) > (la->end - la->tail)) {
		logdbg(stderr, "enqueue: rewind tail to %p\n", la->tail);
			la->tail = la->start;
	}

	/* not enough space on head : drop msg */
	if (la->head > la->tail &&
	    (len + sizeof(struct logmsg)) > (la->head - la->tail)) {
		logdbg(stderr, "enqueue: log area overrun, drop msg\n");

		if (!la->empty)
			la->tail = lastmsg;

		return 1;
	}

	/* ok, we can stage the msg in the area */
	la->empty = 0;
	msg = (struct logmsg *)la->tail;
	msg->prio = prio;
	memcpy((void *)&msg->str, buff, len);
	lastmsg->next = la->tail;
	msg->next = la->head;

	logdbg(stderr, "enqueue: %p, %p, %i, %s\n", (void *)msg, msg->next,
		msg->prio, (char *)&msg->str);

#if LOGDBG
	dump_logarea();
#endif
	return 0;
}

static int log_dequeue(void *buff)
{
	struct logmsg * src = (struct logmsg *)la->head;
	struct logmsg * dst = (struct logmsg *)buff;
	struct logmsg * lst = (struct logmsg *)la->tail;

	if (la->empty)
		return 1;

	int len = strlen((char *)&src->str) * sizeof(char) +
		  sizeof(struct logmsg) + 1;

	dst->prio = src->prio;
	memcpy(dst, src,  len);

	if (la->tail == la->head)
		la->empty = 1; /* we purge the last logmsg */
	else {
		la->head = src->next;
		lst->next = la->head;
	}
	logdbg(stderr, "dequeue: %p, %p, %i, %s\n",
		(void *)src, src->next, src->prio, (char *)&src->str);

	memset((void *)src, 0,  len);

	return la->empty;
}

/*
 * this one can block under memory pressure
 */
static void log_syslog (void * buff)
{
	struct logmsg * msg = (struct logmsg *)buff;
	char* str = (char *)&msg->str;
	size_t width = strlen(str);
	char buf[MAX_MSG_SIZE];

	if (use_syslog)
		syslog(msg->prio, "%s", str);
	if (use_pcslog) {
		if (width && (str[width-1]=='\n')) {
			strcpy(buf,str);
			buf[width-1]=0;
			str = buf;
		}
		tgt_pcs_log(XLOG_WARN,"%s", str);
	}
}

static void dolog(int prio, const char *fmt, va_list ap)
{
	struct timespec ts;
	struct sembuf ops;

	if (la) {
		ts.tv_sec = 0;
		ts.tv_nsec = 10000000;

		ops.sem_num = 0;
		ops.sem_flg = 0;
		ops.sem_op = -1;
		if (semtimedop(la->semid, &ops, 1, &ts) < 0) {
			syslog(LOG_ERR, "semop up failed");
			return;
		}

		log_enqueue(prio, fmt, ap);

		ops.sem_op = 1;
		if (semop(la->semid, &ops, 1) < 0) {
			syslog(LOG_ERR, "semop down failed");
			return;
		}
	} else {
		fprintf(stderr, "%s: ", log_name?log_name:"tgtd");
		vfprintf(stderr, fmt, ap);
		fflush(stderr);
	}
}

void log_warning(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	dolog(LOG_WARNING, fmt, ap);
	va_end(ap);
}

void log_error(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	dolog(LOG_ERR, fmt, ap);
	va_end(ap);
}

void log_debug(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	dolog(LOG_DEBUG, fmt, ap);
	va_end(ap);
}

static void log_flush(void)
{
	struct sembuf ops;

	if (!la)
		return;

	while (!la->empty) {
		ops.sem_num = 0;
		ops.sem_flg = 0;
		ops.sem_op = -1;
		if (semop(la->semid, &ops, 1) < 0) {
			syslog(LOG_ERR, "semop up failed");
			exit(1);
		}

		log_dequeue(la->buff);

		ops.sem_op = 1;
		if (semop(la->semid, &ops, 1) < 0) {
			syslog(LOG_ERR, "semop down failed");
			exit(1);
		}
		log_syslog(la->buff);
	}
}

static void* log_pipe_fn(void* file) {
	FILE *tgt_stderr=file;
	char buf[1024];
	int err;
	while(fgets(buf,sizeof buf, tgt_stderr)) {
		size_t len = strlen(buf);
		if (len && (buf[len-1]=='\n'))
			buf[len-1]=0;
		tgt_pcs_log(XLOG_ERR,"%s(stderr): %s",log_name,buf);
	}
	fclose(tgt_stderr);
	err = la->active;
	if (err)
		tgt_pcs_log(XLOG_ERR,"tgtd exits abnormally, logger pid:%d", getpid());
	log_flush();
	closelog();
	free_logarea();
	exit(err);
}

static pthread_t log_pipe_thread;
static pid_t tgtd_main_pid;

#define LOG_SETUP_ERROR0(fmt) do {		\
		fprintf(stderr,fmt "\n");	\
		syslog(LOG_ERR,fmt);		\
	} while(0);

#define LOG_SETUP_ERROR(fmt,...) do {			\
		fprintf(stderr,fmt "\n", __VA_ARGS__);	\
		syslog(LOG_ERR,fmt,__VA_ARGS__);	\
	} while(0);


int log_init(char *program_name, int size, int daemon, int debug)
{
	const char* logfile = getenv("TGT_LOG_FILE");
	const char* s_logrotate_filenum = getenv("TGT_LOGROTATE_FILENUM");
	const char* s_logrotate_size = getenv("TGT_LOGROTATE_SIZEMB");
	unsigned long logrotate_filenum = s_logrotate_filenum ? strtoul(s_logrotate_filenum, NULL, 0) : 10,
		logrotate_size = s_logrotate_size ? strtoul(s_logrotate_size, NULL, 0) : 100;
	ssize_t nread;
	int status;

	if (logrotate_filenum < 1) logrotate_filenum = 1;
	if (logrotate_size < 1) logrotate_size = 1;

	is_debug = debug;
	tgtd_main_pid = getpid();

	logdbg(stderr,"enter log_init\n");
	log_name = program_name;

	if (!logfile) logfile = "/var/log/tgtd.gz";
	if (daemon) {
		int logstatuspipe[2];
		int logstderrpipe[2];
		FILE *tgt_stderr;
		if (pipe(logstatuspipe)<0) {
			LOG_SETUP_ERROR0("failed to initialize logger status pipe\n");
			goto err_return;
		}
		if (pipe(logstderrpipe)<0) {
			LOG_SETUP_ERROR0("failed to initialize logger stderr pipe\n");
			goto err_close_logstatuspipe;
		}
		openlog(log_name, 0, LOG_DAEMON);
		setlogmask (LOG_UPTO (LOG_DEBUG));

		if (logarea_init(size)) {
			LOG_SETUP_ERROR0("failed to initialize the logger\n");
			goto err_close_logstderrpipe;
		}
		la->active = 1;
		pid = fork();
		if (pid < 0) {
			LOG_SETUP_ERROR0( "fail to fork the logger\n");
			goto err_free_logarea;
		} else if (pid) {
			char st;
			/* in parent process */
			close(logstatuspipe[1]);
			while((nread=read(logstatuspipe[0],&st,1))<0 && (errno==EINTR || errno==EAGAIN));
			close(logstatuspipe[0]);
			close(logstderrpipe[0]);
			if (!nread) {
				return 1;
			} else {
				dup2(logstderrpipe[1], STDERR_FILENO);
				close(logstderrpipe[1]);
				syslog(LOG_WARNING,
				       "tgtd daemon started, pid:%d\n", getpid());
				syslog(LOG_WARNING,
				       "tgtd logger started, pid:%d debug:%d\n", pid, is_debug);
				return 0;
			}
		} /* child */
		close(logstatuspipe[0]);
		close(logstderrpipe[1]);
		tgt_pcs_set_logrotate_filenum(logrotate_filenum);
		tgt_pcs_set_logrotate_size(1024ULL*1024*logrotate_size);

		if (tgt_pcs_set_logfile(logfile)) {
			LOG_SETUP_ERROR("tgtd log file not opened: %s", logfile);
			exit(1);
		}
		tgt_stderr = fdopen(logstderrpipe[0],"r");
		if (!tgt_stderr) {
			char *err = strerror(errno);
			LOG_SETUP_ERROR("fdopen error: %s", err);
			exit(1);
		}

		status = pthread_create(&log_pipe_thread,NULL,log_pipe_fn,tgt_stderr);
		if (status) {
			LOG_SETUP_ERROR0("tgtd logger thread creation error");
			exit(1);
		}

		/* close logger's stderr */
		dup2(STDOUT_FILENO,STDERR_FILENO);
		/* report success */
		while(write(logstatuspipe[1],"",1)<0 && (errno==EINTR || errno==EAGAIN));
		tgt_pcs_log(XLOG_WARN,"started tgtd pid:%d logger pid:%d",tgtd_main_pid,getpid());
		close(logstatuspipe[1]);
		use_pcslog = 1;
		use_syslog = 0;
		while (1) {
			log_flush();
			sleep(10);
		}
	err_free_logarea:
		free_logarea();
	err_close_logstderrpipe:
		close(logstderrpipe[0]);
		close(logstderrpipe[1]);
	err_close_logstatuspipe:
		close(logstatuspipe[1]);
		close(logstatuspipe[0]);
	err_return:
		return 1;
	}
	return 0;
}

void log_close(void)
{
	if (la) {
		log_warning("stopping tgtd logger (pid:%d) for tgtd(pid:%d)\n", pid,tgtd_main_pid);
		syslog(LOG_WARNING,
		       "stopping tgtd logger (pid:%d) for tgtd(pid:%d)\n", pid,tgtd_main_pid);
		la->active = 0;
		/* close original stderr, notifying the logger that
		 * tgtd is exiting (la->active==0 means clean exit) */
		dup2(STDOUT_FILENO,STDERR_FILENO);
		/* wait upon logger to complete */
		waitpid(pid, NULL, 0);
		closelog();
		free_logarea();
	}
}
