#include <stdint.h>
#include <inttypes.h>
#include <time.h>
#include <stdbool.h>
#include <string.h>
#include <assert.h>
#include <sys/resource.h>
#include <sys/timerfd.h>
#include <sys/epoll.h>
#include <fcntl.h>

#include "list.h"
#include "util.h"
#include "tgtd.h"
#include "work.h"
#include "driver.h"
#include "target.h"
#include "scsi.h"


static int timerfd;

uint64_t shaper_cmd_capacity(struct scsi_cmd *cmd, enum tgt_shaper s)
{
	switch (cmd->scb[0]) {
	case WRITE_6:
	case WRITE_10:
	case WRITE_12:
	case WRITE_16:
	case READ_6:
	case READ_10:
	case READ_12:
	case READ_16:
	case WRITE_SAME:
	case WRITE_SAME_16:
		switch (s) {
		case SHAPE_KBPS:
			return scsi_get_in_length(cmd) + scsi_get_out_length(cmd);
		case SHAPE_IOPS:
			return 1;
		default:;
		}
	}
	return 0;
}

int shaper_cmd_enabled(struct scsi_cmd *cmd)
{
	int i;
	uint64_t cap[SHAPE_LIMIT];

	/* check all limits first */
	for (i = 0; i < SHAPE_LIMIT; ++i) {
		cap[i] = shaper_cmd_capacity(cmd, i);
		if (cmd->c_target->shaper[i].cbs != 0 &&
		    cap[i] > cmd->c_target->shaper[i].capacity)
			return 0;

		if (cmd->dev->shaper[i].cbs != 0 &&
		    cap[i] > cmd->dev->shaper[i].capacity)
			return 0;
	}

	/* commit if the command passes */
	for (i = 0; i < SHAPE_LIMIT; ++i) {
		if (cmd->c_target->shaper[i].cbs != 0)
			cmd->c_target->shaper[i].capacity -= cap[i];
		if (cmd->dev->shaper[i].cbs != 0)
			cmd->dev->shaper[i].capacity -= cap[i];
	}
	return 1;
}

void shaper_tick_device(struct scsi_lu* dev)
{
	int i;
	struct timespec ts;

	clock_gettime(CLOCK_MONOTONIC, &ts); /* FIXME */

	for (i = 0; i < SHAPE_LIMIT; ++i) {
		struct tba_shaper *s = &dev->shaper[i];
		uint64_t cbs = s->cbs;

		if (i == SHAPE_KBPS && cbs < dev->attrs.max_tl_bytes) {
			/*
			 * Ensure effective CBS is as big as
			 * max_tl_bytes, to prevent any command from
			 * being stuck in the queue forever.
			 */
			cbs = dev->attrs.max_tl_bytes;
		}
		if (s->capacity < cbs) {
			/* refill and wake the queue with post_done */
			uint64_t elapsed;

			elapsed = ts_valid_diff(&dev->shaper[i].last_tick, &ts);
			s->capacity += (s->out_rate * elapsed)/1000000000ULL;
			if (s->capacity > cbs)
				s->capacity = cbs;
			rescan_device_queue(&dev->cmd_queue);
		}
		s->last_tick = ts;
	}
}

void shaper_tick_target(struct target *target)
{
	int i;
	struct timespec ts;
	struct scsi_lu *dev;

	clock_gettime(CLOCK_MONOTONIC, &ts);

	for (i = 0; i < SHAPE_LIMIT; i++) {
		struct tba_shaper *s = &target->shaper[i];
		uint64_t cbs = s->cbs;
		list_for_each_entry(dev, &target->device_list, device_siblings) {
			if (dev->lun && i == SHAPE_KBPS &&
			    cbs < dev->attrs.max_tl_bytes) {
				/* ensure effective CBS is as big as
				 * max_tl_bytes, to prevent any command from
				 * being stuck in the queue forever.
				 */
				cbs = dev->attrs.max_tl_bytes;
			}
		}

		if (s->capacity < cbs) {
			/* refill and wake the queue with post_done */
			uint64_t elapsed = ts_valid_diff(&s->last_tick, &ts);
			s->capacity += (s->out_rate * elapsed)/1000000000ULL;
			if (s->capacity > cbs)
				s->capacity = cbs;
		}

		list_for_each_entry(dev, &target->device_list, device_siblings)
			if (dev->lun)
				rescan_device_queue(&dev->cmd_queue);
		s->last_tick = ts;
	}
}

void shaper_tick_alldev(void)
{
	int i;
	struct target *tgt;
	struct scsi_lu *dev;

	for (i = 0; tgt_drivers[i]; i++) {
		list_for_each_entry(tgt, &tgt_drivers[i]->target_list, lld_siblings) {
			shaper_tick_target(tgt);
			list_for_each_entry(dev, &tgt->device_list, device_siblings) {
				if (dev->lun) {
					shaper_tick_device(dev);
				}
			}
		}
	}
}

void shaper_handle_tick(int fd, int events, void *data)
{
	char buf[8];

	read(fd, buf, sizeof buf);
	shaper_tick_alldev();
}

int shaper_init(void)
{
	static const struct itimerspec its = {
		.it_interval = {0, 100 * 1000 * 1000},
		.it_value = {0, 100 * 1000 * 1000},
	};

	timerfd = timerfd_create(CLOCK_MONOTONIC, 0);
	if (timerfd < 0) {
		eprintf("timerfd creation failed: %m\n");
		return -1;
	}
	tgt_event_add(timerfd, EPOLLIN, shaper_handle_tick, NULL);
	/* set interval */
	timerfd_settime(timerfd, 0, &its, NULL);
	return 0;
}

void shaper_close(void)
{
	close(timerfd);
}
